/*
 * Copyright (C) 2010	Andre Beckedorf
 * 				 		<evilJazz _AT_ katastrophos _DOT_ net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <cstdlib>
#include <ctime>
#include <time.h>

#include <qstring.h>
#include <qapplication.h>
#include <qdatetime.h>

#include <QHeaderView>
#include <QScrollBar>

#include <QMenu>

#include <QPainter>
#include <QImage>

#include "mediadatabase.h"

#include "media.h"
#include "playlistview.h"
#include "playlistaspectlistviewmodel.h"
#include "playbackcontroller.h"
#include "debug.h"
#include "skinmanager.h"
#include "sqlite3.h"

#include "compathack.h"

/* PlayListItemDelegate */

PlayListItemDelegate::PlayListItemDelegate(PlayListView *playListView)
    :	QStyledItemDelegate(playListView),
        playListView_(playListView)
{

}

void PlayListItemDelegate::paint(QPainter *p, const QStyleOptionViewItem &option, const QModelIndex &index) const
{
    QStyledItemDelegate::paint(p, option, index);

    if (index.column() == 0 && playListView_->playbackController_)
    {
        int ownHeight = option.rect.height();
        int r = 1;

        if (index.row() == playListView_->playbackController_->activeItemIndex())
        {
            const QPixmap *indicator = !playListView_->playbackController_->isPaused() ? playListView_->playIndicatorImage_ : playListView_->pauseIndicatorImage_;
            if (indicator)
            {
                int t = option.rect.top() + (ownHeight - indicator->height()) / 2;
                p->drawPixmap(r, t, *indicator);
                r += indicator->width();
            }
        }

        PlayListItem *item = playListView_->aspectSource()->itemList()->at(index.row());

        if (item->media()->errorDetected() && playListView_->errorIndicatorImage_)
        {
            const QPixmap *indicator = playListView_->errorIndicatorImage_;
            int t = option.rect.top() + (ownHeight - indicator->height()) / 2;
            p->drawPixmap(r, t, *indicator);
        }
    }
}

/* PlayListView */

PlayListView::PlayListView(QWidget *parent)
    :	QTreeView(parent),
        aspectSource_(NULL),
        playbackController_(NULL),
        delegate_(this),
        updateCount_(0),
        userIsResizingSection_(false),
        autoResizingSections_(false),
        evenRowColorGroup_(),
        oddRowColorGroup_(),
        playIndicatorImage_(NULL),
        pauseIndicatorImage_(NULL),
        errorIndicatorImage_(NULL),
        inputMode_(Select),
        itemOrderDirty_(false)
{
    DGUARDMETHODTIMED;
    setFrameStyle(QFrame::NoFrame);

    setItemDelegate(&delegate_);

    setItemsExpandable(false);
    setSortingEnabled(true);
    setUniformRowHeights(true);
    setRootIsDecorated(false);
    setWordWrap(false);
    setAllColumnsShowFocus(true);

    setInputMode(PlayListView::Select);

    connect(this, SIGNAL(doubleClicked(const QModelIndex &)), this, SIGNAL(playbackIntent()));
    //connect(this, SIGNAL(activated(const QModelIndex &)), this, SIGNAL(playbackIntent()));
    //connect(this, SIGNAL(clicked(const QModelIndex &)), this, SLOT(itemClicked()));

    setAlternatingRowColors(true);
    setAutoScaleColumnsEnabled(false);

    setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);
}

PlayListView::~PlayListView()
{

}

void PlayListView::attachTo(PlayListAspectListViewRenderer *source, PlaybackController *controller)
{
    DGUARDMETHODTIMED;
    detach();

    if (source && controller)
    {
        aspectSource_ = source;
        playbackController_ = controller;

        setModel(aspectSource_->model());

        int width = header()->width();

        header()->setCascadingSectionResizes(true);
        header()->setStretchLastSection(false);

        header()->setResizeMode(0, QHeaderView::Fixed);
        header()->resizeSection(0, 20);
        header()->setResizeMode(1, QHeaderView::Fixed);
        header()->resizeSection(1, 60);
        header()->setResizeMode(2, QHeaderView::Stretch);
        header()->setResizeMode(3, QHeaderView::Stretch);
        header()->setResizeMode(4, QHeaderView::Stretch);
        header()->setResizeMode(5, QHeaderView::Fixed);
        header()->resizeSection(5, 80);

        connect(aspectSource_, SIGNAL(updateView()), this, SLOT(updateView()));
        connect(aspectSource_, SIGNAL(graphExecutionStarted(bool &)), this, SLOT(graphExecutionStarted(bool &)));

        connect(this, SIGNAL(playbackIntent()), playbackController_, SLOT(stopCurrentPlaybackAndPlayCurrentItem()));
        connect(playbackController_, SIGNAL(activeItemChanged(int, int)), this, SLOT(playbackControllerActiveItemChanged(int, int)));
        connect(playbackController_, SIGNAL(playbackPaused()), this, SLOT(playbackControllerPlaybackStatusChanged()));
        connect(playbackController_, SIGNAL(playbackUnpaused()), this, SLOT(playbackControllerPlaybackStatusChanged()));
        connect(playbackController_, SIGNAL(playbackStarted()), this, SLOT(playbackControllerPlaybackStatusChanged()));
        connect(playbackController_, SIGNAL(startingPlayback()), this, SLOT(playbackControllerPlaybackStatusChanged()));

        header()->setSortIndicator(aspectSource_->sortedColumn(), !aspectSource_->isSortAscending() ? Qt::AscendingOrder : Qt::DescendingOrder);

        updateView();

/*
        setCurrentItem(findItemByIndex(aspectSource_->selectedItemIndex()));
        setContentsPos(aspectSource_->viewportPosition().x(), aspectSource_->viewportPosition().y());
*/
    }
}

void PlayListView::detach()
{
    DGUARDMETHODTIMED;
    if (aspectSource_ && playbackController_)
    {
        ensurePlayListOrderIsValid();

/*
        aspectSource_->setSelectedItemIndex(currentItem() ? itemIndex(currentItem()) : -1);
        aspectSource_->setViewportPosition(QPoint(contentsX(), contentsY()));
*/
        disconnect(playbackController_, SIGNAL(startingPlayback()), this, SLOT(playbackControllerPlaybackStatusChanged()));
        disconnect(playbackController_, SIGNAL(playbackStarted()), this, SLOT(playbackControllerPlaybackStatusChanged()));
        disconnect(playbackController_, SIGNAL(playbackUnpaused()), this, SLOT(playbackControllerPlaybackStatusChanged()));
        disconnect(playbackController_, SIGNAL(playbackPaused()), this, SLOT(playbackControllerPlaybackStatusChanged()));
        disconnect(playbackController_, SIGNAL(activeItemChanged(int, int)), this, SLOT(playbackControllerActiveItemChanged(int, int)));
        disconnect(this, SIGNAL(playbackIntent()), playbackController_, SLOT(stopCurrentPlaybackAndPlayCurrentItem()));

        disconnect(aspectSource_, SIGNAL(graphExecutionStarted(bool &)), this, SLOT(graphExecutionStarted(bool &)));
        disconnect(aspectSource_, SIGNAL(updateView()), this, SLOT(updateView()));
/*
        activeItem_ = NULL;
        selectedItems_.clear();
        QListView::clear();
*/
        itemOrderDirty_ = false;

        setModel(NULL);

        playbackController_ = NULL;
        aspectSource_ = NULL;
    }
}

PlayListItem *PlayListView::currentPlayListItem()
{
    if (currentIndex().isValid() && aspectSource_ && currentIndex().row() < aspectSource_->itemList()->count())
    {
        return aspectSource_->itemList()->at(currentIndex().row());
    }
    else
    {
        return NULL;
    }
}

int PlayListView::selectedItemsCount()
{
    return currentIndex().isValid() ? 1 : 0;
}

void PlayListView::playbackControllerActiveItemChanged(int newActiveItemIndex, int oldActiveItemIndex)
{
    if (oldActiveItemIndex > -1)
    {
        QModelIndex activeItem = model()->index(oldActiveItemIndex, 0);
        update(activeItem);
    }

    if (newActiveItemIndex > -1)
    {
        QModelIndex activeItem = model()->index(newActiveItemIndex, 0);
        scrollTo(activeItem, EnsureVisible);
        update(activeItem);
    }
}

void PlayListView::playbackControllerPlaybackStatusChanged()
{
    if (playbackController_->activeItemIndex() > -1)
    {
        QModelIndex activeItem = model()->index(playbackController_->activeItemIndex(), 0);
        update(activeItem);
    }
}

void PlayListView::selectionChanged(const QItemSelection &selected, const QItemSelection &deselected)
{
    QTreeView::selectionChanged(selected, deselected);
    //updateSelectedPlayListItem();
    //emit currentIndexChanged();
}

void PlayListView::currentChanged(const QModelIndex &current, const QModelIndex &previous)
{
    QTreeView::currentChanged(current, previous);
    updateSelectedPlayListItem();
    emit currentIndexChanged();
}

void PlayListView::itemClicked()
{
    //updateSelectedPlayListItem();
    //emit currentIndexChanged();
}

void PlayListView::updateSelectedPlayListItem()
{
    if (aspectSource_)
    {
        if (currentIndex().isValid())
            aspectSource_->setSelectedItemIndex(currentIndex().row());
        else
            aspectSource_->setSelectedItemIndex(-1);
    }
}

void PlayListView::beginUpdate()
{
    DENTERMETHOD;

    if (updateCount_ == 0)
    {
        viewport()->setUpdatesEnabled(false);
        setUpdatesEnabled(false);
        blockSignals(true);
    }
    ++updateCount_;

    DEXITMETHOD;
}

void PlayListView::endUpdate()
{
    DENTERMETHOD;

    --updateCount_;
    if (updateCount_ == 0)
    {
        DTIMERINIT(time);
        blockSignals(false);
        setUpdatesEnabled(true);
        viewport()->setUpdatesEnabled(true);
//		triggerUpdate();
        DTIMERPRINT(time, "list view refresh");
    }

    DEXITMETHOD;
}

void PlayListView::ensurePlayListOrderIsValid()
{
/*
    if (!aspectSource_) return;

    DENTERMETHOD;
    if (!aspectSource_->isGraphFiltered() && itemOrderDirty_)
    {
        DPRINTF("playlist item order is dirty, we have to save the order to the DB...");
        beginUpdate();

        PlayListViewItem *item;
        PlayListItemList itemList;
#ifndef QT4
        itemList.setAutoDelete(false);
#endif
        QListViewItemIterator it(this);
        while (it.current())
        {
            item = static_cast<PlayListViewItem *>(it.current());
            itemList.append(item->item_);
            ++it;
        }

        if (itemList.count() > 0)
            aspectSource_->setItemList(itemList);

        itemOrderDirty_ = false;
        itemList.clear();

        endUpdate();
    }
    DEXITMETHOD;
*/
}

void PlayListView::graphExecutionStarted(bool &stopIt)
{
    DENTERMETHOD;

    if (itemOrderDirty_)
    {
        ensurePlayListOrderIsValid();
        stopIt = true;
    }

    DEXITMETHOD;
}

void PlayListView::updateView()
{
    if (aspectSource()) static_cast<PlayListAspectListViewModel *>(aspectSource()->model())->rereadData();

/*
    DENTERMETHOD;

    beginUpdate();

    QListView::clear();

    header()->setSortIndicator(aspectSource_->sortedColumn(), !aspectSource_->isSortAscending());

    PlayListViewItem *lastNewItem = NULL;
    PlayListItemList *items = aspectSource_->itemList();

    activeItem_ = NULL;
    PlayListViewItem *selectedItem = NULL;

    for (int i = 0; i < items->count(); ++i)
    {
        lastNewItem = new PlayListViewItem(items->at(i), i, this, lastNewItem);

        if (!activeItem_ && i == playbackController_->activeItemIndex())
            activeItem_ = lastNewItem;

        if (!selectedItem && i == aspectSource_->selectedItemIndex())
            selectedItem = lastNewItem;
    }

    if (selectedItem)
    {
        setCurrentItem(selectedItem);
        ensureItemVisible(selectedItem);
    }

    endUpdate();

    if (autoScaleColumnsEnabled_)
        setAutoSectionSizes();

    DEXITMETHOD;
*/
}

void PlayListView::setSortedColumn(PlayListView::Columns column)
{
    sortByColumn(column, Qt::AscendingOrder);
    updateView();
}

void PlayListView::setInputMode(PlayListView::InputMode inputmode)
{
/*
    switch (inputmode)
    {
    case PlayListView::Select:
        selectAll(false);
        setSelectionMode(QListView::Single);
        if (currentItem())
            currentItem()->setSelected(true);
        break;
    case PlayListView::MultiSelect:
    case PlayListView::Move:
        setSelectionMode(QListView::Extended);
        break;
    }

    inputMode_ = inputmode;
*/
}

MediaIDList PlayListView::getSelectedMediaIDs()
{
    MediaIDList list;
    if (aspectSource())
    {
        const QModelIndexList selectedModelIndexes = selectedIndexes();
        for (int i = 0; i < selectedModelIndexes.count(); ++i)
        {
            PlayListItem *item = aspectSource()->itemList()->at(selectedModelIndexes.at(i).row());
            list.append(item->mediaID());
        }
    }
    return list;
}

MediaList PlayListView::getSelectedMedia()
{
    MediaList list;
    if (aspectSource())
    {
        const QModelIndexList selectedModelIndexes = selectedIndexes();
        for (int i = 0; i < selectedModelIndexes.count(); ++i)
        {
            PlayListItem *item = aspectSource()->itemList()->at(selectedModelIndexes.at(i).row());
            list.append(item->media());
        }
    }
    return list;
}

PlayListIDList PlayListView::getSelectedItemIDs()
{
    PlayListIDList list;
    if (aspectSource())
    {
        const QModelIndexList selectedModelIndexes = selectedIndexes();
        for (int i = 0; i < selectedModelIndexes.count(); ++i)
        {
            PlayListItem *item = aspectSource()->itemList()->at(selectedModelIndexes.at(i).row());
            list.append(item->id());
        }
    }
    return list;
}

void PlayListView::removeSelectedItems()
{
/*
    if (selectedItems_.count() > 0)
    {
        if (itemOrderDirty_)
        {
            for (int i = selectedItems_.count() - 1; i > -1; --i)
            {
                PlayListViewItem *item = selectedItems_.at(i);
                selectedItems_.removeAt(i);
                delete item;
            }
        }
        else
        {
            // important to get the selected IDs here, since beginChangingUpdate
            // will clear the list and thus the selection...
            PlayListIDList itemIDs = getSelectedItemIDs();
            aspectSource_->playListAspect()->playList()->deletePlayListIDs(itemIDs);
        }
    }
*/
}

void PlayListView::makeItemVisibleByID(unsigned long playListID, bool selectItem)
{
    if (aspectSource())
    {
        int index = aspectSource()->findIndexByPlayListID(playListID);
        if (index > -1)
        {
            QModelIndex modelIndex = aspectSource()->model()->index(index);
            scrollTo(modelIndex, EnsureVisible);

            if (selectItem)
                setCurrentIndex(modelIndex);
        }
    }
}

void PlayListView::insertMediaIDsAtCurrentIndex(const MediaIDList &srcMediaIDs)
{
/*
    // if the playlist is not sorted or filtered, then we can
    // directly insert the new items at the current position...
    if (!aspectSource()->isGraphFiltered())
    {
        PlayListViewItem *lastPlayListViewItem = NULL;
        bool moveFirstItem = false;

        // is anything selected?
        if (selectedItems().count() > 0)
        {
            // if yes, take the first item as place to insert...
            lastPlayListViewItem = selectedItems().at(0);
            DHOP(lastPlayListViewItem->media()->dumpInfo());

            // since we want to place the new items at the position/index of
            // the previously selected insertion point, we need to get the item
            // above it, because we can only tell QListView to insert after
            // a given item...
            if (lastPlayListViewItem->itemAbove())
                lastPlayListViewItem = static_cast<PlayListViewItem *>(lastPlayListViewItem->itemAbove());
            else if (lastPlayListViewItem == firstItem())
                // if there is no item above, this means we have selected the first
                // child in the list. Do a sanity check and then mark the item separately
                // so we can move it to the end of the new stuff after our items have
                // been inserted...
                moveFirstItem = true;
        }
        else
            // insert at end of list if nothing is selected or the playlist is empty...
            lastPlayListViewItem = lastItem();

        selectAll(false);

        PlayListItemList newItems;
        aspectSource_->playListAspect()->playList()->appendMediaIDs(srcMediaIDs, &newItems);

        for (PlayListItemList::ConstIterator listit = newItems.begin(); listit != newItems.end(); ++listit)
        {
            lastPlayListViewItem = new PlayListViewItem((*listit), 0, this, lastPlayListViewItem);
            setSelected(lastPlayListViewItem, true);
        }

        // if we have a marked an item that was (and still is) the first child in the list
        // move it to the end of our inserted items...
        if (moveFirstItem)
            firstItem()->moveItem(lastPlayListViewItem);

        itemOrderDirty_ = true;
    }
    // the playlist is sorted or filtered, so insert at DB level...
    else
        aspectSource_->playListAspect()->playList()->appendMediaIDs(srcMediaIDs);
*/
}

void PlayListView::setAutoScaleColumnsEnabled(bool value)
{
    autoScaleColumnsEnabled_ = value;
}

void PlayListView::saveColumnOrder(const QString &configName, const QString &groupName)
{
    DGUARDMETHODTIMED;
    Config cfg(configName, Config::File);
    cfg.setGroup(groupName);

    for (int i = 0; i < header()->length(); ++i)
    {
        int logicalIndex = header()->logicalIndex(i);
        cfg.writeEntry("section_" + QString::number(i), QString::number(logicalIndex));
    }
}

void PlayListView::saveColumnWidth(const QString &configName, const QString &groupName)
{
    DGUARDMETHODTIMED;
    Config cfg(configName, Config::File);
    cfg.setGroup(groupName);

    for (int i = 0; i < header()->length(); ++i)
        cfg.writeEntry("width_" + QString::number(i), header()->sectionSize(i));
}

void PlayListView::loadColumnOrder(const QString &configName, const QString &groupName)
{
    DGUARDMETHODTIMED;
    Config cfg(configName, Config::File);
    cfg.setGroup(groupName);

    for (int i = 0; i < header()->length(); ++i)
    {
        QString key = "section_" + QString::number(i);
        if (cfg.hasKey(key))
        {
            QStringList args = cfg.readListEntry(key, ',');
            int logicalIndex = args[0].toInt();
            header()->moveSection(header()->visualIndex(logicalIndex), i);
        }
    }
}

void PlayListView::loadColumnWidth(const QString &configName, const QString &groupName)
{
    DGUARDMETHODTIMED;
    Config cfg(configName, Config::File);
    cfg.setGroup(groupName);

    for (int i = 0; i < header()->length(); ++i)
    {
        QString key = "width_" + QString::number(i);
        if (cfg.hasKey(key))
            header()->resizeSection(i, qMax(10, cfg.readNumEntry(key)));
    }

    setColumnWidth(0, 20);
}

void PlayListView::setBackgroundColors(const QColor &evenColor, const QColor &oddColor)
{
    oddRowColorGroup_.setColor(QPalette::Base, oddColor);
    evenRowColorGroup_.setColor(QPalette::Base, evenColor);
}

void PlayListView::setForegroundColor(const QColor &textColor)
{
    oddRowColorGroup_.setColor(QPalette::Text, textColor);
    oddRowColorGroup_.setColor(QPalette::Foreground, textColor);
    evenRowColorGroup_.setColor(QPalette::Text, textColor);
    evenRowColorGroup_.setColor(QPalette::Foreground, textColor);
}

void PlayListView::setPlayIndicatorImage(const QPixmap &image)
{
    playIndicatorImage_ = &image;
}

void PlayListView::setPauseIndicatorImage(const QPixmap &image)
{
    pauseIndicatorImage_ = &image;
}

void PlayListView::setErrorIndicatorImage(const QPixmap &image)
{
    errorIndicatorImage_ = &image;
}

void PlayListView::setSkinMode(const SkinModeInformation &modeInfo)
{
    if (modeInfo.playListFont().font.isUsed())
        setFont(modeInfo.playListFont().font);
    else
        setFont(QFont());

    // set general palette:
    QPalette palette = QApplication::palette();

    if (modeInfo.playListFont().fontColor.isUsed())
    {
        palette.setColor(QPalette::Text, modeInfo.playListFont().fontColor);
        palette.setColor(QPalette::Foreground, modeInfo.playListFont().fontColor);
    }

    if (modeInfo.playListBackgroundColorSelected().isUsed())
    {
        palette.setColor(QPalette::Highlight, modeInfo.playListBackgroundColorSelected());
        palette.setColor(QPalette::HighlightedText, modeInfo.playListFontColorSelected());
    }

    // set even and odd colors for items:
    if (modeInfo.playListBackgroundColorEven().isUsed())
    {
        if (modeInfo.playListBackgroundColorOdd().isUsed())
            setBackgroundColors(modeInfo.playListBackgroundColorEven(), modeInfo.playListBackgroundColorOdd());
        else
            setBackgroundColors(modeInfo.playListBackgroundColorEven(), modeInfo.playListBackgroundColorEven());

        palette.setColor(QPalette::Base, modeInfo.playListBackgroundColorEven());

        if (modeInfo.playListBackgroundColorOdd().isUsed())
            palette.setColor(QPalette::AlternateBase, modeInfo.playListBackgroundColorOdd());
    }
    else
        setBackgroundColors(palette.color(QPalette::Active, QPalette::Base), palette.color(QPalette::Active, QPalette::Base));

    setPalette(palette);

    setPlayIndicatorImage(modeInfo.playListPlayIndicatorImage());
    setPauseIndicatorImage(modeInfo.playListPauseIndicatorImage());
    setErrorIndicatorImage(modeInfo.playListErrorIndicatorImage());

    if (modeInfo.playListHeaderFont().font.isUsed())
        header()->setFont(modeInfo.playListHeaderFont().font);
    else
        header()->setFont(QFont());

    if (modeInfo.playListHeaderColor().isUsed())
    {
        QPalette palette(modeInfo.playListHeaderColor());

        if (modeInfo.playListHeaderFont().fontColor.isUsed())
        {
            palette.setColor(QPalette::Text, modeInfo.playListHeaderFont().fontColor);
            palette.setColor(QPalette::ButtonText, modeInfo.playListHeaderFont().fontColor);
            palette.setColor(QPalette::Foreground, modeInfo.playListHeaderFont().fontColor);
        }

        header()->setPalette(palette);
    }
    else
        header()->setPalette(QPalette());

    if (modeInfo.playListScrollBarColor().isUsed())
    {
        horizontalScrollBar()->setPalette(QPalette(modeInfo.playListScrollBarColor()));
        verticalScrollBar()->setPalette(QPalette(modeInfo.playListScrollBarColor()));
    }
    else
    {
        horizontalScrollBar()->setPalette(QPalette());
        verticalScrollBar()->setPalette(QPalette());
    }
}

void PlayListView::mousePressEvent(QMouseEvent *e)
{
    QTreeView::mousePressEvent(e);

    if (e->button() == Qt::RightButton && popupMenu_)
        popupMenu_->exec(e->globalPos());
}
